@@ -1,6 +1,5 @@
 pub mod colibri;
 mod conference;
-mod connection;
 mod jingle;
 mod pinger;
 mod source;
@@ -9,11 +8,12 @@ mod util;
 mod xmpp;
 
 pub use crate::{
-  colibri::ColibriMessage,
-  conference::{JitsiConference, JitsiConferenceConfig, Participant},
-  connection::JitsiConnection,
+  conference::{Feature, JitsiConference, JitsiConferenceConfig, Participant},
   source::MediaType,
+  stanza_filter::StanzaFilter,
+  xmpp::connection::{Authentication, Connection},
 };
+pub use xmpp_parsers;
 
 #[cfg(feature = "tracing-subscriber")]
 pub fn init_tracing(level: tracing::Level) {
@@ -22,4 +22,4 @@ pub fn init_tracing(level: tracing::Level) {
     .with_span_events(tracing_subscriber::fmt::format::FmtSpan::CLOSE)
     .with_target(false)
     .init();
-}
\ No newline at end of file
+}
