@@ -3,7 +3,7 @@ use async_trait::async_trait;
 use xmpp_parsers::Element;
 
 #[async_trait]
-pub(crate) trait StanzaFilter {
+pub trait StanzaFilter {
   fn filter(&self, element: &Element) -> bool;
   async fn take(&self, element: Element) -> Result<()>;
 }
